"""Abstractions for creating argparse parsers from a dataclass definition."""

from __future__ import annotations

import argparse
import collections.abc
import dataclasses
import inspect
import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

import termcolor
import typing_extensions
from typing_extensions import get_args, get_origin

from . import _arguments, _docstrings, _fields, _instantiators, _resolver, _strings

T = TypeVar("T")


_known_parsable_types = set(
    filter(
        lambda x: isinstance(x, Hashable),  # type: ignore
        itertools.chain(
            __builtins__.values(),  # type: ignore
            vars(typing_extensions).values(),
            vars(collections.abc).values(),
        ),
    )
)


def _is_possibly_nested_type(typ: Any) -> bool:
    """Heuristics for determining whether a type can be treated as a 'nested type',
    where a single field has multiple corresponding argumentsi (eg for nested
    dataclasses or classes).

    Examples of when we return False: int, str, List[int], List[str], pathlib.Path, etc.
    """

    origin = get_origin(typ)
    if origin is not None:
        typ = origin

    # Nested types should be callable.
    if not callable(typ):
        return False

    # Known parsable types: builtins like int/str/float/bool, collections, annotations.
    if typ in _known_parsable_types:
        return False

    # Simple heuristic: dataclasses should be treated as nested objects and are not
    # parsable.
    if dataclasses.is_dataclass(typ):
        return True

    # Non-parsable types like nested (data)classes should have fully type-annotated
    # inputs. If any inputs are unannotated (for example, in the case of pathlib.Path),
    # we can assume the type is parsable.
    for param in inspect.signature(typ).parameters.values():
        if param.annotation is inspect.Parameter.empty:
            return False

    return True


@dataclasses.dataclass(frozen=True)
class ParserSpecification:
    """Each parser contains a list of arguments and optionally some subparsers."""

    description: str
    args: List[_arguments.ArgumentDefinition]
    helptext_from_nested_class_field_name: Dict[str, Optional[str]]
    subparsers: Optional[SubparsersSpecification]

    @staticmethod
    def from_callable(
        f: Callable[..., T],
        description: Optional[str],
        parent_classes: Set[Type],
        parent_type_from_typevar: Optional[Dict[TypeVar, Type]],
        default_instance: Optional[T],
    ) -> ParserSpecification:
        """Create a parser definition from a callable."""

        # Resolve generic types.
        f, type_from_typevar = _resolver.resolve_generic_types(f)
        if parent_type_from_typevar is not None:
            for typevar, typ in type_from_typevar.items():
                if typ in parent_type_from_typevar:
                    type_from_typevar[typevar] = parent_type_from_typevar[typ]  # type: ignore

        # Cycle detection.
        if f in parent_classes:
            raise _instantiators.UnsupportedTypeAnnotationError(
                f"Found a cyclic dataclass dependency with type {f}."
            )

        # TODO: we are abusing the (minor) distinctions between types, classes, and
        # callables throughout the code. This is mostly for legacy reasons, could be
        # cleaned up.
        parent_classes = parent_classes | {cast(Type, f)}

        args = []
        helptext_from_nested_class_field_name = {}
        subparsers = None
        field_list = _fields.field_list_from_callable(
            f=f, default_instance=default_instance
        )
        for field in field_list:

            field = dataclasses.replace(
                field,
                typ=type_from_typevar.get(field.typ, field.typ),  # type: ignore
            )
            if isinstance(field.typ, TypeVar):
                # Found an unbound TypeVar. This could be because inheriting from
                # generics is currently not implemented. It's unclear whether this is
                # feasible, because generics are lost in the mro:
                # https://github.com/python/typing/issues/777
                raise _instantiators.UnsupportedTypeAnnotationError(
                    f"Field {field.name} has an unbound TypeVar: {field.typ}. Note that"
                    " inheriting from generics is currently not implemented. It's"
                    " unclear whether this is feasible, because generics are lost in"
                    " the mro: https://github.com/python/typing/issues/777"
                )

            # (1) Handle Unions over callables; these result in subparsers.
            subparsers_attempt = SubparsersSpecification.from_field(
                field,
                type_from_typevar=type_from_typevar,
                parent_classes=parent_classes,
            )
            if subparsers_attempt is not None:
                if subparsers is not None:
                    raise _instantiators.UnsupportedTypeAnnotationError(
                        "Only one set of subparsers is allowed."
                    )
                subparsers = subparsers_attempt
                continue

            # (2) Handle nested callables.
            if _is_possibly_nested_type(field.typ):
                nested_parser = ParserSpecification.from_callable(
                    field.typ,
                    description=None,
                    parent_classes=parent_classes,
                    parent_type_from_typevar=type_from_typevar,
                    default_instance=field.default,
                )
                for arg in nested_parser.args:
                    args.append(
                        dataclasses.replace(
                            arg,
                            prefix=field.name
                            + _strings.NESTED_DATACLASS_DELIMETER
                            + arg.prefix,
                        )
                    )
                for k, v in nested_parser.helptext_from_nested_class_field_name.items():
                    helptext_from_nested_class_field_name[
                        field.name + _strings.NESTED_DATACLASS_DELIMETER + k
                    ] = v

                if field.helptext is not None:
                    helptext_from_nested_class_field_name[field.name] = field.helptext
                else:
                    helptext_from_nested_class_field_name[field.name] = inspect.getdoc(
                        _resolver.resolve_generic_types(field.typ)[0]
                    )
                continue

            # (3) Handle primitive types. These produce a single argument!
            args.append(
                _arguments.ArgumentDefinition.from_field(field, type_from_typevar)
            )

        return ParserSpecification(
            description=description
            if description is not None
            else _docstrings.get_callable_description(f),
            args=args,
            helptext_from_nested_class_field_name=helptext_from_nested_class_field_name,
            subparsers=subparsers,
        )

    def apply(self, parser: argparse.ArgumentParser) -> None:
        """Create defined arguments and subparsers."""

        parser.description = self.description

        def format_group_name(nested_field_name: str, required: bool) -> str:
            if required:
                prefix = termcolor.colored("required", attrs=["bold"])
            else:
                prefix = termcolor.colored("optional", attrs=["bold", "dark"])
            suffix = termcolor.colored("arguments", attrs=["bold"])

            if nested_field_name != "":
                return " ".join(
                    [
                        prefix,
                        termcolor.colored(nested_field_name, attrs=["bold"]),
                        suffix,
                    ]
                )
            else:
                return " ".join([prefix, suffix])

        optional_group_from_prefix: Dict[str, argparse._ArgumentGroup] = {
            "": parser._action_groups[1],
        }
        required_group_from_prefix: Dict[str, argparse._ArgumentGroup] = {
            "": parser.add_argument_group(format_group_name("", required=True)),
        }

        # Break some API boundaries to rename the optional group.
        parser._action_groups[1].title = format_group_name("", required=False)
        positional_group = parser.add_argument_group(
            termcolor.colored("positional arguments", attrs=["bold"])
        )
        parser._action_groups = parser._action_groups[::-1]

        # Add each argument.
        for arg in self.args:
            if arg.field.positional:
                arg.add_argument(positional_group)
                continue

            if arg.required:
                target_groups, other_groups = (
                    required_group_from_prefix,
                    optional_group_from_prefix,
                )
            else:
                target_groups, other_groups = (
                    optional_group_from_prefix,
                    required_group_from_prefix,
                )

            if arg.prefix not in target_groups:
                nested_field_name = arg.prefix[:-1]
                target_groups[arg.prefix] = parser.add_argument_group(
                    format_group_name(nested_field_name, required=arg.required),
                    # Add a description, but only to the first group for a field.
                    description=self.helptext_from_nested_class_field_name[
                        nested_field_name
                    ]
                    if arg.prefix not in other_groups
                    else None,
                )
            arg.add_argument(target_groups[arg.prefix])

        # Add subparsers.
        if self.subparsers is not None:
            title = "subcommands"
            metavar = "{" + ",".join(self.subparsers.parser_from_name.keys()) + "}"
            if not self.subparsers.required:
                title = "optional " + title
                metavar = f"[{metavar}]"

            argparse_subparsers = parser.add_subparsers(
                dest=_strings.SUBPARSER_DEST_FMT.format(name=self.subparsers.name),
                description=self.subparsers.description,
                required=self.subparsers.required,
                title=title,
                metavar=metavar,
            )
            for name, subparser_def in self.subparsers.parser_from_name.items():
                subparser = argparse_subparsers.add_parser(name)
                subparser_def.apply(subparser)


@dataclasses.dataclass(frozen=True)
class SubparsersSpecification:
    """Structure for defining subparsers. Each subparser is a parser with a name."""

    name: str
    description: Optional[str]
    parser_from_name: Dict[str, ParserSpecification]
    required: bool
    default_instance: Optional[Any]

    @staticmethod
    def from_field(
        field: _fields.Field,
        type_from_typevar: Dict[TypeVar, Type],
        parent_classes: Set[Type],
    ) -> Optional[SubparsersSpecification]:
        # Union of classes should create subparsers.
        if get_origin(field.typ) is not Union:
            return None

        # We don't use sets here to retain order of subcommands.
        options = [type_from_typevar.get(typ, typ) for typ in get_args(field.typ)]
        options_no_none = [o for o in options if o != type(None)]  # noqa
        if len(options_no_none) < 2 or not all(
            map(_is_possibly_nested_type, options_no_none)
        ):
            return None

        parser_from_name: Dict[str, ParserSpecification] = {}
        for option in options_no_none:
            subparser_name = _strings.subparser_name_from_type(option)
            parser_from_name[subparser_name] = ParserSpecification.from_callable(
                option,
                description=None,
                parent_classes=parent_classes,
                parent_type_from_typevar=type_from_typevar,
                default_instance=None,
            )

        return SubparsersSpecification(
            name=field.name,
            # If we wanted, we could add information about the default instance
            # automatically, as is done for normal fields. But for now we just rely on
            # the user to include it in the docstring.
            description=field.helptext,
            parser_from_name=parser_from_name,
            # Required if: type hint is not Optional[], or a default instance is
            # provided.
            required=(options == options_no_none) and field.default is None,
            default_instance=field.default,
        )



"""Type-safe, human-readable serialization helpers."""

import dataclasses
import datetime
import enum
from typing import IO, Any, Optional, Set, Type, TypeVar, Union

import yaml
from typing_extensions import get_origin

from . import _resolver

ENUM_YAML_TAG_PREFIX = "!enum:"
DATACLASS_YAML_TAG_PREFIX = "!dataclass:"

DataclassType = TypeVar("DataclassType")


def _get_contained_special_types_from_instance(instance: Any) -> Set[Type]:
    """Takes an object and recursively searches its cihldren for dataclass or enum
    types."""
    if issubclass(type(instance), enum.Enum):
        return {type(instance)}
    elif not dataclasses.is_dataclass(instance):
        return set()

    out = {type(instance)}
    for v in vars(instance).values():
        out |= _get_contained_special_types_from_instance(v)
    return out


def _get_contained_special_types_from_type(
    cls: Type,
    _parent_contained_dataclasses: Optional[Set[Type]] = None,
) -> Set[Type]:
    """Takes a dataclass type, and recursively searches its fields for dataclass or enum
    types."""
    assert _resolver.is_dataclass(cls)
    parent_contained_dataclasses = (
        set()
        if _parent_contained_dataclasses is None
        else _parent_contained_dataclasses
    )

    cls, type_from_typevar = _resolver.resolve_generic_types(cls)

    contained_dataclasses = {cls}

    def handle_type(typ) -> Set[Type]:
        if _resolver.is_dataclass(typ) and typ not in parent_contained_dataclasses:
            return _get_contained_special_types_from_type(
                typ,
                _parent_contained_dataclasses=contained_dataclasses
                | parent_contained_dataclasses,
            )
        elif type(typ) is enum.EnumMeta:
            return {typ}
        return set()

    # Handle generics.
    for typ in type_from_typevar.values():
        contained_dataclasses |= handle_type(typ)

    if cls in parent_contained_dataclasses:
        return contained_dataclasses

    # Handle fields.
    for field in _resolver.resolved_fields(cls):  # type: ignore
        contained_dataclasses |= handle_type(field.type)

    return contained_dataclasses


def _make_loader(cls: Type) -> Type[yaml.Loader]:
    class DataclassLoader(yaml.Loader):
        pass

    # Design Q: do we want to support multiple dataclass types with the same name?
    # - Why yes: rudimentary support for this is easy.
    # - Why no: might cause confusing error messages? Using all possible context cues
    # for this is also hard.
    #
    # => let's just keep things simple, assert uniqueness for now. Easier to add new
    # features later than remove them.

    contained_types = list(_get_contained_special_types_from_type(cls))
    contained_type_names = list(map(lambda cls: cls.__name__, contained_types))
    assert len(set(contained_type_names)) == len(contained_type_names), (
        "Contained dataclass type names must all be unique, but got"
        f" {contained_type_names}"
    )

    loader: yaml.Loader
    node: yaml.Node

    def make_dataclass_constructor(typ: Type):
        return lambda loader, node: typ(**loader.construct_mapping(node))

    def make_enum_constructor(typ: Type):
        return lambda loader, node: typ[loader.construct_python_str(node)]

    for typ, name in zip(contained_types, contained_type_names):
        if dataclasses.is_dataclass(typ):
            DataclassLoader.add_constructor(
                tag=DATACLASS_YAML_TAG_PREFIX + name,
                constructor=make_dataclass_constructor(typ),
            )
        elif issubclass(typ, enum.Enum):
            DataclassLoader.add_constructor(
                tag=ENUM_YAML_TAG_PREFIX + name,
                constructor=make_enum_constructor(typ),
            )
        else:
            assert False

    return DataclassLoader


def _make_dumper(instance: Any) -> Type[yaml.Dumper]:
    class DataclassDumper(yaml.Dumper):
        pass

    contained_types = list(_get_contained_special_types_from_instance(instance))
    contained_type_names = list(map(lambda cls: cls.__name__, contained_types))

    # Note: this is currently a stricter than necessary assert.
    assert len(set(contained_type_names)) == len(contained_type_names), (
        "Contained dataclass/enum names must all be unique, but got"
        f" {contained_type_names}"
    )

    dumper: yaml.Dumper
    data: Any
    field: dataclasses.Field

    def make_representer(name: str):
        def representer(dumper, data):
            if dataclasses.is_dataclass(data):
                return dumper.represent_mapping(
                    tag=DATACLASS_YAML_TAG_PREFIX + name,
                    mapping={
                        field.name: getattr(data, field.name)
                        for field in dataclasses.fields(data)
                        if field.init
                    },
                )
            elif isinstance(data, enum.Enum):
                return dumper.represent_scalar(
                    tag=ENUM_YAML_TAG_PREFIX + name, value=data.name
                )

        return representer

    for typ, name in zip(contained_types, contained_type_names):
        DataclassDumper.add_representer(typ, make_representer(name))
    return DataclassDumper


def from_yaml(
    cls: Type[DataclassType],
    stream: Union[str, IO[str], bytes, IO[bytes]],
) -> DataclassType:
    """Re-construct a dataclass instance from a yaml-compatible string, which should be
    generated from `dcargs.to_yaml()`."""
    out = yaml.load(stream, Loader=_make_loader(cls))
    origin_cls = get_origin(cls)
    assert isinstance(out, origin_cls if origin_cls is not None else cls)
    return out


def _timestamp() -> str:
    """Get a current timestamp as a string. Example format: `2021-11-05-15:46:32`."""
    return datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")


def to_yaml(instance: Any) -> str:
    """Serialize a dataclass; returns a yaml-compatible string that can be deserialized
    via `dcargs.from_yaml()`."""
    return f"# YAML generated via dcargs, at {_timestamp()}.\n" + yaml.dump(
        instance, Dumper=_make_dumper(instance)
    )



"""Helpers for parsing dataclass docstrings. Used for helptext generation."""
import dataclasses
import functools
import inspect
import io
import tokenize
from typing import Callable, Dict, List, Optional, Type

import docstring_parser
from typing_extensions import get_origin

from . import _resolver, _strings


@dataclasses.dataclass(frozen=True)
class _Token:
    token_type: int
    content: str
    logical_line: int
    actual_line: int


@dataclasses.dataclass(frozen=True)
class _FieldData:
    index: int
    logical_line: int
    actual_line: int
    prev_field_logical_line: int


@dataclasses.dataclass(frozen=True)
class _ClassTokenization:
    tokens: List[_Token]
    tokens_from_logical_line: Dict[int, List[_Token]]
    tokens_from_actual_line: Dict[int, List[_Token]]
    field_data_from_name: Dict[str, _FieldData]

    @staticmethod
    @functools.lru_cache(maxsize=8)
    def make(clz) -> "_ClassTokenization":
        """Parse the source code of a class, and cache some tokenization information."""
        readline = io.BytesIO(inspect.getsource(clz).encode("utf-8")).readline

        tokens: List[_Token] = []
        tokens_from_logical_line: Dict[int, List[_Token]] = {1: []}
        tokens_from_actual_line: Dict[int, List[_Token]] = {1: []}
        field_data_from_name: Dict[str, _FieldData] = {}

        logical_line: int = 1
        actual_line: int = 1
        for toktype, tok, start, end, line in tokenize.tokenize(readline):
            # Note: we only track logical line numbers, which are delimited by
            # `tokenize.NEWLINE`. `tokenize.NL` tokens appear when logical lines are
            # broken into multiple lines of code; these are ignored.
            if toktype == tokenize.NEWLINE:
                logical_line += 1
                actual_line += 1
                tokens_from_logical_line[logical_line] = []
                tokens_from_actual_line[actual_line] = []
            elif toktype == tokenize.NL:
                actual_line += 1
                tokens_from_actual_line[actual_line] = []
            elif toktype is not tokenize.INDENT:
                token = _Token(
                    token_type=toktype,
                    content=tok,
                    logical_line=logical_line,
                    actual_line=actual_line,
                )
                tokens.append(token)
                tokens_from_logical_line[logical_line].append(token)
                tokens_from_actual_line[actual_line].append(token)

        prev_field_logical_line: int = 1
        for i, token in enumerate(tokens[:-1]):
            if token.token_type == tokenize.NAME:
                # Naive heuristic for field names.
                if (
                    tokens[i + 1].content == ":"
                    and tokens[i] == tokens_from_actual_line[tokens[i].actual_line][0]
                    and token.content not in field_data_from_name
                ):
                    field_data_from_name[token.content] = _FieldData(
                        index=i,
                        logical_line=token.logical_line,
                        actual_line=token.actual_line,
                        prev_field_logical_line=prev_field_logical_line,
                    )
                    prev_field_logical_line = token.logical_line

        return _ClassTokenization(
            tokens=tokens,
            tokens_from_logical_line=tokens_from_logical_line,
            tokens_from_actual_line=tokens_from_actual_line,
            field_data_from_name=field_data_from_name,
        )


def get_class_tokenization_with_field(
    cls: Type, field_name: str
) -> Optional[_ClassTokenization]:
    # Search for token in this class + all parents.
    found_field: bool = False
    classes_to_search = cls.mro()
    tokenization = None
    for search_cls in classes_to_search:
        # Inherited generics seem challenging for now.
        # https://github.com/python/typing/issues/777
        assert get_origin(search_cls) is None

        try:
            tokenization = _ClassTokenization.make(search_cls)  # type: ignore
        except OSError as e:
            # Dynamic dataclasses will result in an OSError -- this is fine, we just assume
            # there's no docstring.
            assert "could not find class definition" in e.args[0]
            return None
        except TypeError as e:  # pragma: no cover
            # Notebooks cause “___ is a built-in class” TypeError.
            assert "built-in class" in e.args[0]
            return None

        # Grab field-specific tokenization data.
        if field_name in tokenization.field_data_from_name:
            found_field = True
            break

    if dataclasses.is_dataclass(cls):
        assert found_field, (
            "Docstring parsing error -- this usually means that there are multiple"
            " dataclasses in the same file with the same name but different scopes."
        )

    return tokenization


def get_field_docstring(cls: Type, field_name: str) -> Optional[str]:
    """Get docstring for a field in a class."""

    tokenization = get_class_tokenization_with_field(cls, field_name)
    if tokenization is None:  # Currently only happens for dynamic dataclasses.
        return None

    field_data = tokenization.field_data_from_name[field_name]

    # Check for docstring-style comment. This should be on the next logical line.
    logical_line = field_data.logical_line + 1
    if (
        logical_line in tokenization.tokens_from_logical_line
        and len(tokenization.tokens_from_logical_line[logical_line]) >= 1
    ):
        first_token = tokenization.tokens_from_logical_line[logical_line][0]
        first_token_content = first_token.content.strip()

        # Found a docstring!
        if (
            first_token.token_type == tokenize.STRING
            and first_token_content.startswith('"""')
            and first_token_content.endswith('"""')
        ):
            return _strings.dedent(first_token_content[3:-3])

    # Check for comment on the same line as the field.
    final_token_on_line = tokenization.tokens_from_logical_line[
        field_data.logical_line
    ][-1]
    if final_token_on_line.token_type == tokenize.COMMENT:
        comment: str = final_token_on_line.content
        assert comment.startswith("#")
        return comment[1:].strip()

    # Check for comments that come before the field. This is intentionally written to
    # support comments covering multiple (grouped) fields, for example:
    #
    #     # Optimizer hyperparameters.
    #     learning_rate: float
    #     beta1: float
    #     beta2: float
    #
    # In this case, 'Optimizer hyperparameters' will be treated as the docstring for all
    # 3 fields.
    comments: List[str] = []
    current_actual_line = field_data.actual_line - 1
    while current_actual_line in tokenization.tokens_from_actual_line:
        actual_line_tokens = tokenization.tokens_from_actual_line[current_actual_line]
        current_actual_line -= 1

        # We stop looking if we find an empty line.
        if len(actual_line_tokens) == 0:
            break

        # Record single comments!
        if (
            len(actual_line_tokens) == 1
            and actual_line_tokens[0].token_type == tokenize.COMMENT
        ):
            (comment_token,) = actual_line_tokens
            assert comment_token.content.startswith("#")
            comments.append(comment_token.content[1:].strip())
        elif len(comments) > 0:
            # Comments should be contiguous.
            break

    if len(comments) > 0:
        return "\n".join(reversed(comments))

    return None


def get_callable_description(f: Callable) -> str:
    """Get description associated with a callable via docstring parsing.

    Note that the `dataclasses.dataclass` will automatically populate __doc__ based on
    the fields of the class if a docstring is not specified; this helper will ignore
    these docstrings."""
    f, _unused = _resolver.resolve_generic_types(f)

    # Ignore any default docstrings, as generated by `dataclasses.dataclass`.
    docstring = inspect.getdoc(f)
    if docstring is None:
        return ""

    if dataclasses.is_dataclass(f):
        default_doc = f.__name__ + str(inspect.signature(f)).replace(" -> None", "")
        if docstring == default_doc:
            return ""
        return docstring
    else:
        parsed_docstring = docstring_parser.parse(docstring)
        return "\n".join(
            list(
                filter(
                    lambda x: x is not None,  # type: ignore
                    [
                        parsed_docstring.short_description,
                        parsed_docstring.long_description,
                    ],
                )
            )
        )



from ._cli import cli, parse
from ._instantiators import UnsupportedTypeAnnotationError
from ._serialization import from_yaml, to_yaml

__all__ = [
    "cli",
    # Deprecated.
    # "parse",
    "UnsupportedTypeAnnotationError",
    "from_yaml",
    "to_yaml",
]



"""Utilities for working with strings."""
import enum
import functools
import re
import textwrap
from typing import Type, TypeVar

from typing_extensions import get_args

from . import _resolver

NESTED_DATACLASS_DELIMETER: str = "."
SUBPARSER_DEST_FMT: str = "{name} (positional)"


def dedent(text: str) -> str:
    """Same as textwrap.dedent, but ignores the first line."""
    first_line, line_break, rest = text.partition("\n")
    if line_break == "":
        return textwrap.dedent(text)
    return f"{first_line.strip()}\n{textwrap.dedent(rest)}"


_camel_separator_pattern = functools.lru_cache(maxsize=1)(
    lambda: re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
)


def hyphen_separated_from_camel_case(name: str) -> str:
    return _camel_separator_pattern().sub(r"-\1", name).lower()


def subparser_name_from_type(cls: Type) -> str:
    cls, type_from_typevar = _resolver.resolve_generic_types(cls)
    if len(type_from_typevar) == 0:
        assert hasattr(cls, "__name__")
        return hyphen_separated_from_camel_case(cls.__name__)  # type: ignore

    return "-".join(
        map(
            subparser_name_from_type,
            [cls] + list(type_from_typevar.values()),
        )
    )


T = TypeVar("T")


def instance_from_string(typ: Type[T], arg: str) -> T:
    """Given a type and and a string from the command-line, reconstruct an object. Not
    intended to deal with containers.

    This is intended to replace all calls to `type(string)`, which can cause unexpected
    behavior. As an example, note that the following argparse code will always print
    `True`, because `bool("True") == bool("False") == bool("0") == True`.
    ```
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--flag", type=bool)

    print(parser.parse_args().flag)
    ```
    """
    assert len(get_args(typ)) == 0, f"Type {typ} cannot be instantiated."
    if typ is bool:
        return {"True": True, "False": False}[arg]  # type: ignore
    elif issubclass(typ, enum.Enum):
        return typ[arg]  # type: ignore
    else:
        return typ(arg)  # type: ignore



"""Utilities for resolving generic types and forward references."""

import copy
import dataclasses
from typing import Callable, Dict, List, Tuple, Type, TypeVar, Union

from typing_extensions import get_args, get_origin, get_type_hints


def is_dataclass(cls: Union[Type, Callable]) -> bool:
    """Same as `dataclasses.is_dataclass`, but also handles generic aliases."""
    origin_cls = get_origin(cls)
    return dataclasses.is_dataclass(cls if origin_cls is None else origin_cls)


TypeOrCallable = TypeVar("TypeOrCallable", Type, Callable)


def resolve_generic_types(
    cls: TypeOrCallable,
) -> Tuple[TypeOrCallable, Dict[TypeVar, Type]]:
    """If the input is a class: no-op. If it's a generic alias: returns the origin
    class, and a mapping from typevars to concrete types."""

    origin_cls = get_origin(cls)
    if origin_cls is not None:
        typevars = origin_cls.__parameters__
        typevar_values = get_args(cls)
        assert len(typevars) == len(typevar_values)
        return origin_cls, dict(zip(typevars, typevar_values))
    else:
        return cls, {}


def resolved_fields(cls: Type) -> List[dataclasses.Field]:
    """Similar to dataclasses.fields, but resolves forward references."""

    assert dataclasses.is_dataclass(cls)
    fields = []
    annotations = get_type_hints(cls)
    for field in dataclasses.fields(cls):
        # Avoid mutating original field.
        field = copy.copy(field)

        # Resolve forward references.
        field.type = annotations[field.name]

        fields.append(field)

    return fields



"""Core functionality for calling functions with arguments specified by argparse
namespaces."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Set, Tuple, TypeVar

from typing_extensions import get_args

from . import _fields, _parsers, _resolver, _strings


class InstantiationError(Exception):
    """Exception raised when instantiation fail; this typically means that values from
    the CLI are invalid."""


T = TypeVar("T")


def call_from_args(
    f: Callable[..., T],
    parser_definition: _parsers.ParserSpecification,
    value_from_arg: Dict[str, Any],
    field_name_prefix: str = "",
) -> Tuple[T, Set[str]]:
    """Call `f` with arguments specified by a dictionary of values from argparse.

    Returns the output of `f` and a set of used arguments."""

    f, type_from_typevar = _resolver.resolve_generic_types(f)

    args: List[str] = []
    kwargs: Dict[str, Any] = {}
    consumed_keywords: Set[str] = set()

    def get_value_from_arg(arg: str) -> Any:
        """Helper for getting values from `value_from_arg` + doing some extra
        asserts."""
        assert arg in value_from_arg
        assert arg not in consumed_keywords
        consumed_keywords.add(arg)
        return value_from_arg[arg]

    arg_from_prefixed_field_name = {}
    for arg in parser_definition.args:
        arg_from_prefixed_field_name[arg.prefix + arg.field.name] = arg

    for field in _fields.field_list_from_callable(f, default_instance=None):  # type: ignore
        value: Any
        prefixed_field_name = field_name_prefix + field.name

        # Resolve field type.
        field_type = (
            type_from_typevar[field.typ]  # type: ignore
            if field.typ in type_from_typevar
            else field.typ
        )

        if prefixed_field_name in arg_from_prefixed_field_name:
            # Standard arguments.
            arg = arg_from_prefixed_field_name[prefixed_field_name]
            value = get_value_from_arg(prefixed_field_name)
            if value is not None:
                try:
                    assert arg.instantiator is not None
                    value = arg.instantiator(value)
                except ValueError as e:
                    raise InstantiationError(
                        f"Parsing error for {arg.get_name_or_flag()}: {e.args[0]}"
                    )
        elif (
            prefixed_field_name
            in parser_definition.helptext_from_nested_class_field_name
        ):
            # Nested callable.
            value, consumed_keywords_child = call_from_args(
                field_type,
                parser_definition,
                value_from_arg,
                field_name_prefix=prefixed_field_name
                + _strings.NESTED_DATACLASS_DELIMETER,
            )
            consumed_keywords |= consumed_keywords_child
        else:
            # Unions over dataclasses (subparsers). This is the only other option.
            assert parser_definition.subparsers is not None
            assert field.name == parser_definition.subparsers.name

            subparser_dest = _strings.SUBPARSER_DEST_FMT.format(
                name=prefixed_field_name
            )
            subparser_name = get_value_from_arg(subparser_dest)
            if subparser_name is None:
                # No subparser selected -- this should only happen when we do either
                # Optional[Union[A, B, ...]] or Union[A, B, None], or have a
                # default/default_factory set.
                assert (
                    type(None) in get_args(field_type)
                    or parser_definition.subparsers.default_instance is not None
                )
                value = parser_definition.subparsers.default_instance
            else:
                options = map(
                    lambda x: x if x not in type_from_typevar else type_from_typevar[x],
                    get_args(field_type),
                )
                chosen_f = None
                for option in options:
                    if _strings.subparser_name_from_type(option) == subparser_name:
                        chosen_f = option
                        break
                assert chosen_f is not None
                value, consumed_keywords_child = call_from_args(
                    chosen_f,
                    parser_definition.subparsers.parser_from_name[subparser_name],
                    value_from_arg,
                )
                consumed_keywords |= consumed_keywords_child

        if field.positional:
            args.append(value)
        else:
            kwargs[field.name] = value

    return f(*args, **kwargs), consumed_keywords  # type: ignore



"""Abstractions for pulling out 'field' abstractions, which specify inputs, from
general callables."""
import dataclasses
import inspect
import warnings
from typing import Any, Callable, List, Optional, Type, TypeVar

import docstring_parser
from typing_extensions import get_type_hints

from . import _docstrings, _resolver


@dataclasses.dataclass(frozen=True)
class Field:
    name: str
    typ: Type
    default: Any
    helptext: Optional[str]
    positional: bool


T = TypeVar("T")


def field_list_from_callable(
    f: Callable[..., T], default_instance: Optional[T]
) -> List[Field]:
    """Generate a list of generic 'field' objects corresponding to an input callable.

    `f` can be from a dataclass type, regular class type, or function."""

    # Unwrap generics.
    f, _unused_type_from_typevar = _resolver.resolve_generic_types(f)

    # Handling for class inputs vs function inputs.
    ignore_self = False
    cls: Optional[Type] = None
    if isinstance(f, type):
        # If `f` is a type:
        #     1. Set cls to the type.
        #     2. Consider `f` to be `cls.__init__`.
        cls = f
        f = cls.__init__  # type: ignore
        ignore_self = True

    # Get type annotations, docstrings.
    hints = get_type_hints(f)
    docstring = inspect.getdoc(f)
    docstring_from_arg_name = {}
    if docstring is not None:
        for param_doc in docstring_parser.parse(docstring).params:
            docstring_from_arg_name[param_doc.arg_name] = param_doc.description
    del docstring

    # Special handling for dataclasses: take `field(default_factory=...)` and
    # `default_instance` input into acocunt.
    default_from_dataclass_field_name = {}
    if cls is not None and _resolver.is_dataclass(cls):
        for field in _resolver.resolved_fields(cls):
            default_from_dataclass_field_name[
                field.name
            ] = _get_dataclass_field_default(field, default_instance)
    else:
        assert (
            default_instance is None
        ), "`default_instance` is only supported for dataclass types."

    # Generate field list from function signature.
    field_list = []
    for param in inspect.signature(f).parameters.values():
        # For `__init__`, skip self parameter.
        if ignore_self:
            ignore_self = False
            continue

        # Get default value.
        default = default_from_dataclass_field_name.get(param.name, param.default)
        if default is inspect.Parameter.empty:
            default = None

        # Get helptext from docstring.
        helptext = docstring_from_arg_name.get(param.name)
        if cls is not None:
            helptext = _docstrings.get_field_docstring(cls, param.name)

        field_list.append(
            Field(
                name=param.name,
                # Note that param.annotation does not resolve forward references.
                typ=hints[param.name],
                default=default,
                helptext=helptext,
                positional=param.kind is inspect.Parameter.POSITIONAL_ONLY,
            )
        )
    return field_list


def _ensure_dataclass_instance_used_as_default_is_frozen(
    field: dataclasses.Field, default_instance: Any
) -> None:
    """Ensure that a dataclass type used directly as a default value is marked as
    frozen."""
    assert dataclasses.is_dataclass(default_instance)
    cls = type(default_instance)
    if not cls.__dataclass_params__.frozen:
        warnings.warn(
            f"Mutable type {cls} is used as a default value for `{field.name}`. This is"
            " dangerous! Consider using `dataclasses.field(default_factory=...)` or"
            f" marking {cls} as frozen."
        )


def _get_dataclass_field_default(
    field: dataclasses.Field, parent_default_instance: Any
) -> Optional[Any]:
    """Helper for getting the default instance for a field."""
    field_default_instance = None
    if field.default is not dataclasses.MISSING:
        # Populate default from usual default value, or
        # `dataclasses.field(default=...)`.
        field_default_instance = field.default
        if dataclasses.is_dataclass(field_default_instance):
            _ensure_dataclass_instance_used_as_default_is_frozen(
                field, field_default_instance
            )
    elif field.default_factory is not dataclasses.MISSING and not (
        # Special case to ignore default_factory if we write:
        # `field: Dataclass = dataclasses.field(default_factory=Dataclass)`.
        #
        # In other words, treat it the same way as:
        # `field: Dataclass`.
        #
        # The only time this matters is when we our dataclass has a `__post_init__`
        # function that mutates the dataclass. We choose here to use the default values
        # before this method is called.
        dataclasses.is_dataclass(field.type)
        and field.default_factory is field.type
    ):
        # Populate default from `dataclasses.field(default_factory=...)`.
        field_default_instance = field.default_factory()

    if parent_default_instance is not None:
        # Populate default from some parent, eg `default_instance` in `dcargs.cli()`.
        field_default_instance = getattr(parent_default_instance, field.name)
    return field_default_instance



"""Helper for using type annotations to recursively generate instantiator functions,
which map strings (or, in some cases, sequences of strings) to the annotated type.

Some examples of type annotations and the desired instantiators:
```
    int

        lambda string: int(str)

    List[int]

        lambda strings: list(
            [int(x) for x in strings]
        )

    List[Color], where Color is an enum

        lambda strings: list(
            [Color[x] for x in strings]
        )

    Tuple[int, float]

        lambda strings: tuple(
            [
                typ(x)
                for typ, x in zip(
                    (int, float),
                    strings,
                )
            ]
        )
```
"""

import collections
import dataclasses
import enum
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Annotated, Final, Literal, get_args, get_origin

from . import _strings

Instantiator = Union[
    # Most standard fields: these are converted from strings from the CLI.
    Callable[[str], Any],
    # Sequence fields! This should be used whenever argparse's `nargs` field is set.
    Callable[[List[str]], Any],
    # Special case: the only time that argparse doesn't give us a string is when the
    # argument action is set to `store_true` or `store_false`. In this case, we get
    # a bool directly, and the field action can be a no-op.
    Callable[[bool], bool],
]


@dataclasses.dataclass
class InstantiatorMetadata:
    nargs: Optional[Union[str, int]]
    metavar: Union[str, Tuple[str, ...]]
    choices: Optional[Tuple[Any, ...]]
    is_optional: bool


class UnsupportedTypeAnnotationError(Exception):
    """Exception raised when an unsupported type annotation is detected."""


_builtin_set = set(
    filter(
        lambda x: isinstance(x, Hashable),  # type: ignore
        __builtins__.values(),  # type: ignore
    )
)


def instantiator_from_type(
    typ: Type, type_from_typevar: Dict[TypeVar, Type]
) -> Tuple[Instantiator, InstantiatorMetadata]:
    """Recursive helper for parsing type annotations.

    Returns two things:
    - An instantiator function, for instantiating the type from a string or list of
      strings. The latter applies when argparse's `nargs` parameter is set.
    - A metadata structure, which specifies parameters for argparse.
    """

    # Resolve typevars.
    if typ in type_from_typevar:
        return instantiator_from_type(
            type_from_typevar[typ],  # type: ignore
            type_from_typevar,
        )

    # Address container types. If a matching container is found, this will recursively
    # call instantiator_from_type().
    container_out = _instantiator_from_container_type(typ, type_from_typevar)
    if container_out is not None:
        return container_out

    # Validate that typ is a `(arg: str) -> T` type converter, as expected by argparse.
    if typ in _builtin_set:
        pass
    elif not callable(typ):
        raise UnsupportedTypeAnnotationError(
            f"Expected {typ} to be a `(arg: str) -> T` type converter, but is not"
            " callable."
        )
    else:
        param_count = 0
        has_var_positional = False
        signature = inspect.signature(typ)
        for i, param in enumerate(signature.parameters.values()):
            if i == 0 and param.annotation not in (str, inspect.Parameter.empty):
                raise UnsupportedTypeAnnotationError(
                    f"Expected {typ} to be a `(arg: str) -> T` type converter, but got"
                    f" {signature}. You may have a nested type in a container, which is"
                    " unsupported."
                )
            if param.kind is inspect.Parameter.VAR_POSITIONAL:
                has_var_positional = True
            elif param.default is inspect.Parameter.empty and param.kind is not (
                inspect.Parameter.VAR_KEYWORD
            ):
                param_count += 1

        # Raise an error if parameters look wrong.
        if not (param_count == 1 or (param_count == 0 and has_var_positional)):
            raise UnsupportedTypeAnnotationError(
                f"Expected {typ} to be a `(arg: str) -> T` type converter, but got"
                f" {signature}. You may have a nested type in a container, which is"
                " unsupported."
            )

    # Special case `choices` for some types, as implemented in `instance_from_string()`.
    auto_choices: Optional[Tuple[str, ...]] = None
    if typ is bool:
        auto_choices = ("True", "False")
    elif issubclass(typ, enum.Enum):
        auto_choices = tuple(x.name for x in typ)

    return lambda arg: _strings.instance_from_string(typ, arg), InstantiatorMetadata(
        nargs=None,
        metavar=typ.__name__.upper(),
        choices=auto_choices,
        is_optional=False,
    )


def _instantiator_from_container_type(
    typ: Type, type_from_typevar: Dict[TypeVar, Type]
) -> Optional[Tuple[Instantiator, InstantiatorMetadata]]:
    """Attempt to create an instantiator from a container type. Returns `None` is no
    container type is found."""

    type_origin = get_origin(typ)
    if type_origin is None:
        return None

    # Unwrap Annotated and Final types.
    if type_origin in (Annotated, Final):
        contained_type = get_args(typ)[0]
        return instantiator_from_type(contained_type, type_from_typevar)

    # List, tuples, and sequences.
    if type_origin in (
        collections.abc.Sequence,  # different from typing.Sequence!
        list,  # different from typing.List!
        set,  # different from typing.Set!
    ):
        (contained_type,) = get_args(typ)
        container_type = type_origin
        if container_type is collections.abc.Sequence:
            container_type = list

        make, inner_meta = _instantiator_from_type_inner(
            contained_type, type_from_typevar
        )
        return lambda strings: container_type(
            [make(x) for x in strings]
        ), InstantiatorMetadata(
            nargs="+",
            metavar=inner_meta.metavar,
            choices=inner_meta.choices,
            is_optional=False,
        )

    # Tuples.
    if type_origin is tuple:
        types = get_args(typ)
        typeset = set(types)  # Note that sets are unordered.
        typeset_no_ellipsis = typeset - {Ellipsis}  # type: ignore

        if typeset_no_ellipsis != typeset:
            # Ellipsis: variable argument counts. When an ellipsis is used, tuples must
            # contain only one type.
            assert len(typeset_no_ellipsis) == 1
            (contained_type,) = typeset_no_ellipsis

            make, inner_meta = _instantiator_from_type_inner(
                contained_type, type_from_typevar
            )
            return lambda strings: tuple(
                [make(x) for x in strings]
            ), InstantiatorMetadata(
                nargs="+",
                metavar=inner_meta.metavar,
                choices=inner_meta.choices,
                is_optional=False,
            )

        else:
            instantiators, metas = zip(
                *map(
                    lambda t: _instantiator_from_type_inner(t, type_from_typevar),
                    types,
                )
            )
            if len(set(m.choices for m in metas)) > 1:
                raise UnsupportedTypeAnnotationError(
                    "Due to constraints in argparse, all choices in fixed-length tuples"
                    " must match. This restricts mixing enums & literals with other"
                    " types."
                )
            return lambda strings: tuple(
                make(x) for make, x in zip(instantiators, strings)
            ), InstantiatorMetadata(
                nargs=len(types),
                metavar=tuple(m.metavar for m in metas),
                choices=metas[0].choices,
                is_optional=False,
            )

    # Optionals.
    if type_origin is Union:
        options = set(get_args(typ))
        if len(options) != 2 or type(None) not in options:
            # Note that the subparsers logic happens much earlier.
            raise UnsupportedTypeAnnotationError(
                "Union must be either over dataclasses (for subparsers) or Optional"
                " (Union[T, None])"
            )
        (typ,) = options - {type(None)}
        instantiator, metadata = _instantiator_from_type_inner(
            typ, type_from_typevar, allow_sequences=True
        )
        return instantiator, dataclasses.replace(metadata, is_optional=True)

    # Literals.
    if type_origin is Literal:
        choices = get_args(typ)
        if not len(set(map(type, choices))) == 1:
            raise UnsupportedTypeAnnotationError(
                "All choices in literal must have the same type!"
            )
        contained_type = type(next(iter(choices)))
        if issubclass(contained_type, enum.Enum):
            choices = tuple(map(lambda x: x.name, choices))
        instantiator, metadata = _instantiator_from_type_inner(
            contained_type, type_from_typevar
        )
        assert (
            # Choices provided by the contained type
            metadata.choices is None
            or len(set(choices) - set(metadata.choices)) == 0
        )
        return instantiator, dataclasses.replace(metadata, choices=choices)

    raise UnsupportedTypeAnnotationError(  # pragma: no cover
        f"Unsupported type {typ} with origin {type_origin}"
    )


def _instantiator_from_type_inner(
    typ: Type,
    type_from_typevar: Dict[TypeVar, Type],
    allow_sequences: bool = False,
    allow_optional: bool = False,
) -> Tuple[Instantiator, InstantiatorMetadata]:
    """Thin wrapper over instantiator_from_type, with some extra asserts for catching
    errors."""
    out = instantiator_from_type(typ, type_from_typevar)
    if not allow_sequences and out[1].nargs is not None:
        raise UnsupportedTypeAnnotationError("Nested sequence types are not supported!")
    if not allow_optional and out[1].is_optional:
        raise UnsupportedTypeAnnotationError("Nested optional types are not supported!")
    return out



import argparse
from typing import Callable, Optional, Sequence, TypeVar

from . import _calling, _parsers, _resolver

T = TypeVar("T")


def parse(
    f: Callable[..., T],
    *,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    default_instance: Optional[T] = None,
) -> T:  # pragma: no cover
    """Deprecated alias for `dcargs.cli()`."""
    # import warnings
    #
    # warnings.warn(
    #     "`dcargs.parse()` has been renamed `dcargs.cli()`. It will be removed"
    #     " soon.",
    #     DeprecationWarning,
    #     stacklevel=2,
    # )
    return cli(f, description=description, args=args, default_instance=default_instance)


def cli(
    f: Callable[..., T],
    *,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    default_instance: Optional[T] = None,
) -> T:
    """Call `f(...)`, with arguments populated from an automatically generated CLI
    interface.

    `f` should have type-annotated inputs, and can be a function, class, or dataclass.
    Note that if `f` is a class, `dcargs.parse()` returns an instance.

    The parser is generated by populating helptext from docstrings and types from
    annotations; a broad range of core type annotations are supported...
        - Types natively accepted by `argparse`: str, int, float, pathlib.Path, etc.
        - Default values for optional parameters.
        - Booleans, which are automatically converted to flags when provided a default
          value.
        - Enums (via `enum.Enum`).
        - Various container types. Some examples:
          - `typing.ClassVar`.
          - `typing.Optional`.
          - `typing.Literal`.
          - `typing.Sequence`.
          - `typing.List`.
          - `typing.Tuple`, such as `typing.Tuple[T1, T2, T3]` or
            `typing.Tuple[T, ...]`.
          - `typing.Set`.
          - `typing.Final` and `typing.Annotated`.
          - Nested combinations of the above: `Optional[Literal[T]]`,
            `Final[Optional[Sequence[T]]]`, etc.
        - Nested dataclasses.
          - Simple nesting.
          - Unions over nested dataclasses (subparsers).
          - Optional unions over nested dataclasses (optional subparsers).
        - Generic dataclasses (including nested generics).

    Args:
        f: Callable.

    Keyword Args:
        description: Description text for the parser, displayed when the --help flag is
            passed in. If not specified, `f`'s docstring is used. Mirrors argument from
            `argparse.ArgumentParser()`.
        args: If set, parse arguments from a sequence of strings instead of the
            commandline. Mirrors argument from `argparse.ArgumentParser.parse_args()`.
        default_instance: An instance of `T` to use for default values; only supported
            if `T` is a dataclass type. Helpful for merging CLI arguments with values loaded
            from elsewhere. (for example, a config object loaded from a yaml file)

    Returns:
        The output of `f(...)`.
    """

    # Map a callable to the relevant CLI arguments + subparsers.
    assert default_instance is None or _resolver.is_dataclass(
        f
    ), "Default instance specification is only supported for dataclasses!"
    parser_definition = _parsers.ParserSpecification.from_callable(
        f,
        description=description,
        parent_classes=set(),  # Used for recursive calls.
        parent_type_from_typevar=None,  # Used for recursive calls.
        default_instance=default_instance,  # Overrides for default values.
    )

    # Parse using argparse!
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser_definition.apply(parser)
    value_from_arg = vars(parser.parse_args(args=args))

    try:
        # Attempt to call `f` using whatever was passed in.
        out, consumed_keywords = _calling.call_from_args(
            f, parser_definition, value_from_arg
        )
    except _calling.InstantiationError as e:
        # Emulate argparse's error behavior when invalid arguments are passed in.
        parser.print_usage()
        print()
        print(e.args[0])
        raise SystemExit()

    assert consumed_keywords == value_from_arg.keys()
    return out



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dcargs",
    version="0.1.3",
    description="Strongly typed, zero-effort CLIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/brentyi/dcargs",
    author="brentyi",
    author_email="brentyi@berkeley.edu",
    license="MIT",
    packages=find_packages(),
    package_data={"dcargs": ["py.typed"]},
    python_requires=">=3.7",
    install_requires=[
        "docstring_parser",
        "typing_extensions>=4.0.0",
        "pyyaml",
        "termcolor",
    ],
    extras_require={
        "testing": [
            "pytest",
            "pytest-cov",
            "attrs",
        ],
        "type-checking": [
            "mypy",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from __future__ import annotations

import argparse
import dataclasses
import enum
import shlex
from typing import Any, Dict, Optional, Set, Tuple, Type, TypeVar, Union

from . import _fields, _instantiators


@dataclasses.dataclass(frozen=True)
class ArgumentDefinition:
    """Options for defining arguments. Contains all necessary arguments for argparse's
    add_argument() method."""

    prefix: str  # Prefix for nesting.
    field: _fields.Field  # Corresponding dataclass field.

    # Action that is called on parsed arguments. This handles conversions from strings
    # to our desired types.
    instantiator: Optional[_instantiators.Instantiator]

    # Fields that will be populated initially.
    # Important: from here on out, all fields correspond 1:1 to inputs to argparse's
    # add_argument() method.
    type: Optional[Union[Type, TypeVar]]
    default: Optional[Any]

    # Fields that will be handled by argument transformations.
    required: bool = False
    action: Optional[str] = None
    nargs: Optional[Union[int, str]] = None
    choices: Optional[Set[Any]] = None
    metavar: Optional[Union[str, Tuple[str, ...]]] = None
    help: Optional[str] = None

    def add_argument(
        self, parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
    ) -> None:
        """Add a defined argument to a parser."""
        kwargs = {k: v for k, v in vars(self).items() if v is not None}

        # Apply prefix for nested dataclasses.
        assert "dest" not in kwargs
        if not self.field.positional:
            kwargs["dest"] = self.prefix + self.field.name

        # Important: as far as argparse is concerned, all inputs are strings.
        #
        # Conversions from strings to our desired types happen in the "field action";
        # this is a bit more flexible, and lets us handle more complex types like enums
        # and multi-type tuples.
        if "type" in kwargs:
            kwargs["type"] = str
        if "choices" in kwargs:
            kwargs["choices"] = list(map(str, kwargs["choices"]))

        kwargs.pop("prefix")
        kwargs.pop("field")
        kwargs.pop("instantiator")

        # Note that the name must be passed in as a position argument.
        parser.add_argument(self.get_name_or_flag(), **kwargs)

    def get_name_or_flag(self) -> str:
        """Get name (for positional args) or flag (for keyword args), with a prefix
        applied for nested dataclasses."""
        if self.field.positional:
            return self.prefix + self.field.name
        elif self.action == "store_false":
            return "--" + (self.prefix + "no-" + self.field.name).replace("_", "-")
        else:
            return "--" + (self.prefix + self.field.name).replace("_", "-")

    @staticmethod
    def from_field(
        field: _fields.Field,
        type_from_typevar: Dict[TypeVar, Type],
    ) -> ArgumentDefinition:
        arg = ArgumentDefinition(
            prefix="",
            field=field,
            instantiator=None,
            type=field.typ,
            default=field.default,
        )
        arg = _transform_required_if_default_set(arg)
        arg = _transform_handle_boolean_flags(arg)
        arg = _transform_recursive_instantiator_from_type(arg, type_from_typevar)
        arg = _transform_generate_helptext(arg)
        arg = _transform_convert_defaults_to_strings(arg)
        arg = _transform_positional_special_handling(arg)
        return arg


def _transform_required_if_default_set(arg: ArgumentDefinition) -> ArgumentDefinition:
    """Set `required=True` if a default value is set."""

    # Mark arg as required if a default is set.
    if arg.default is None:
        return dataclasses.replace(arg, required=True)

    return dataclasses.replace(arg)


def _transform_handle_boolean_flags(arg: ArgumentDefinition) -> ArgumentDefinition:
    """"""
    if arg.type is not bool:
        return arg

    if arg.default is None or arg.field.positional:
        # If no default is passed in, we treat bools as a normal parameter.
        return arg
    elif arg.default is False:
        # Default `False` => --flag passed in flips to `True`.
        return dataclasses.replace(
            arg,
            action="store_true",
            type=None,
            instantiator=lambda x: x,  # argparse will directly give us a bool!
        )
    elif arg.default is True:
        # Default `True` => --no-flag passed in flips to `False`.
        return dataclasses.replace(
            arg,
            action="store_false",
            type=None,
            instantiator=lambda x: x,  # argparse will directly give us a bool!
        )
    else:
        assert False, "Invalid default"


def _transform_recursive_instantiator_from_type(
    arg: ArgumentDefinition,
    type_from_typevar: Dict[TypeVar, Type],
) -> ArgumentDefinition:
    """The bulkiest bit: recursively analyze the type annotation and use it to determine how"""
    if arg.instantiator is not None:
        return arg

    instantiator, metadata = _instantiators.instantiator_from_type(
        arg.type,  # type: ignore
        type_from_typevar,
    )
    return dataclasses.replace(
        arg,
        instantiator=instantiator,
        choices=metadata.choices,
        nargs=metadata.nargs,
        required=(not metadata.is_optional) and arg.required,
        # Ignore metavar if choices is set.
        metavar=metadata.metavar if metadata.choices is None else None,
    )


def _transform_generate_helptext(arg: ArgumentDefinition) -> ArgumentDefinition:
    """Generate helptext from docstring and argument name."""
    help_parts = []

    docstring_help = arg.field.helptext

    if docstring_help is not None and docstring_help != "":
        # Note that the percent symbol needs some extra handling in argparse.
        # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string
        docstring_help = docstring_help.replace("%", "%%")
        help_parts.append(docstring_help)
    elif arg.field.positional:
        help_parts.append(str(arg.metavar))

    if arg.action is not None:
        # Don't show defaults for boolean flags.
        assert arg.action in ("store_true", "store_false")
    elif arg.default is not None and isinstance(arg.default, enum.Enum):
        # Special case for enums.
        help_parts.append(f"(default: {arg.default.name})")
    elif not arg.required:
        # Include default value in helptext. We intentionally don't use the % template
        # because the types of all arguments are set to strings, which will cause the
        # default to be casted to a string and introduce extra quotation marks.
        if arg.nargs is not None and hasattr(arg.default, "__iter__"):
            # For tuple types, we might have arg.default as (0, 1, 2, 3).
            # For list types, we might have arg.default as [0, 1, 2, 3].
            # For set types, we might have arg.default as {0, 1, 2, 3}.
            #
            # In all cases, we want to display (default: 0 1 2 3), for consistency with
            # the format that argparse expects when we set nargs.
            assert arg.default is not None  # Just for type checker.
            default_parts = map(shlex.quote, map(str, arg.default))
            help_parts.append(f"(default: {' '.join(default_parts)})")
        else:
            help_parts.append(f"(default: {shlex.quote(str(arg.default))})")

    return dataclasses.replace(arg, help=" ".join(help_parts))


def _transform_convert_defaults_to_strings(
    arg: ArgumentDefinition,
) -> ArgumentDefinition:
    """Sets all default values to strings, as required as input to our instantiator
    functions. Special-cased for enums."""

    def as_str(x: Any) -> str:
        if isinstance(x, enum.Enum):
            return x.name
        else:
            return str(x)

    if arg.default is None or arg.action is not None:
        return arg
    elif arg.nargs is not None and arg.nargs != "?":
        return dataclasses.replace(arg, default=tuple(map(as_str, arg.default)))
    else:
        return dataclasses.replace(arg, default=as_str(arg.default))


def _transform_positional_special_handling(
    arg: ArgumentDefinition,
) -> ArgumentDefinition:
    """Special handling for positional args."""

    if not arg.field.positional:
        return arg

    return dataclasses.replace(
        arg,
        metavar=(arg.prefix + arg.field.name).upper(),
        required=None,
        nargs="?" if not arg.required else arg.nargs,
    )
